from urllib.parse import unquote
from http.cookies import SimpleCookie
from email.parser import BytesParser as BP
try:
	from orjson import loads
except ModuleNotFoundError:
	try:
		from ujson import loads
	except ModuleNotFoundError:
		from json import loads
from .lib import HTTPStatus


class Request:
	__slots__ = (
			"_headers", "_host", "_method", "_path", "_query_string", "_data",
			"_content_type", "_content_length", "_storage", "cookie", "_ip"
  )

	def __init__(self, req : object):
		self._headers = req
		self._host = req.get("HTTP_HOST", "")
		self._method = req["REQUEST_METHOD"]
		self._path = req["PATH_INFO"]
		self._query_string = None
		self._data = None
		self._content_type = req.get("CONTENT_TYPE")
		self._content_length = int(req.get("CONTENT_LENGTH") or 0)
		self._storage = {}
		self.cookie = SimpleCookie()
		if raw_cookie := req.get("HTTP_COOKIE"):
			self.cookie.load(raw_cookie)
		if req.get("HTTP_X_FORWARDED_FOR"):
			self._ip = req["HTTP_X_FORWARDED_FOR"].split(',')[-1].strip()
		else:
			self._ip = req["REMOTE_ADDR"]

	@property
	def query_string(self) -> dict:
		if self._query_string is not None:
			return self._query_string
		parsed = {}
		for item in unquote(self._headers["QUERY_STRING"]).split("&"):
			pair = item.split("=") + [""]
			if pair[0] in parsed:
				if type(parsed[pair[0]]) == list:
					parsed[pair[0]].append(pair[1])
				else:
					parsed[pair[0]] = [parsed[pair[0]], pair[1]]
			else:
				parsed[pair[0]] = pair[1]
		self._query_string = parsed
		return parsed

	@property
	def data(self) -> dict | list | str | bytes:
		if self._data is not None:
			return self._data
		if not self._content_length:
			self._data = {}
			return self._data
		try:
			if self._content_type[:16] == "application/json":
				self._data = loads(self._headers["wsgi.input"].read(self._content_length))
			elif self._content_type == "application/x-www-form-urlencoded":
				self._data = Request._parse_qsa(unquote(self._headers["wsgi.input"].read(self._content_length).decode()).split("&"))
			elif self._content_type[:20] == "multipart/form-data;":
				m = BP().parsebytes(b"Content-Type: " + self._content_type.encode() + b"\n" + self._headers["wsgi.input"].read(self._content_length))
				self._data = {
					p.get_param("name", header="content-disposition"): p.get_payload(decode=True) for p in m.get_payload()
				}
			else:
				self._data = self._headers["wsgi.input"].read(self._content_length)
			return self._data
		except Exception:
			raise HTTPStatus("400 BAD REQUEST", "400 BAD REQUEST")

	@property
	def headers(self) -> str:
		return self._headers

	@property
	def storage(self) -> dict:
		return self._storage

	@property
	def content_type(self) -> str:
		return self._content_type

	@property
	def content_length(self) -> int:
		return self._content_length

	@property
	def host(self) -> str:
		return self._host

	@property
	def method(self) -> str:
		return self._method

	@property
	def path(self) -> str:
		return self._path

	@property
	def ip(self) -> str:
		return self._ip
