from datetime import datetime
try:
	from orjson import dumps
except ModuleNotFoundError:
	try:
		from ujson import dumps as jdumps
	except ModuleNotFoundError:
		from json import dumps as jdumps
	def dumps(x): return jdumps(x).encode()
from .lib import HTTPStatus


class Response:
	__slots__ = ("_raw", "_body", "_status", "_type", "charset", "_location", "_headers")

	def __init__(self, resp : callable, content_type="application/json", charset="UTF-8"):
		self._raw = resp
		self._status = None
		self._type = content_type
		self.charset = charset
		self._headers = []
		self._body = ""

	def __call__(self) -> bytes:
		if self._type[:4] == "text" or self._type[:11] == "application":
			self.add_header('Content-Type', self._type + ";charset=" + self.charset)
		else:
			self.add_header('Content-Type', self._type)

		if type(self._body) == str:
			body = bytes(self._body, self.charset)
		elif self._type == "application/json":
			body = dumps(self._body)
		elif type(self._body) == bytes:
			body = self._body
		else:
			body = bytes(str(self._body), self.charset)

		self.add_header('Content-Length', str(len(body)))
		self._raw(self._status, self._headers)
		return body


	def set(self, body : str|bytes|dict|list, status : str="", content_type : str=None) -> None:
		self._status = status or "200 OK"
		self._type = content_type or self._type
		self._body = body

	def redirect(self, location : str, status : str="301 MOVED PERMANENTLY") -> None:
		self._status = status
		self._location = location
		self.add_header('Location', location)
		raise HTTPStatus(status)

	def set_cookie(
			self, name : str, value : str, expires : datetime=None,
			maxAge : int|str="", **kwargs) -> None:
		cookie = name + "=" + value
		if type(expires) == datetime:
			cookie += "; Expires=" + expires.strftime('%a, %d %b %Y %H:%M:%S %Z')
		elif expires:
			cookie += "; Expires=" + expires
		elif maxAge:
			cookie += "; Max-Age=" + str(maxAge)
		for arg in kwargs:  # Domain=None,Path=None,Secure=False,HttpOnly=False
			cookie += "; " + arg[0].upper() + arg[1:] + "=" + kwargs[arg] if kwargs[arg] else "; " + arg[0].upper() + arg[1:]
		self._headers.append(("Set-Cookie", cookie))

	def add_header(self, name : str, value : str) -> None:
		self._headers.append((name, value))

	@property
	def body(self) -> str | bytes | dict | list:
		return self._body

	@property
	def status(self) -> str:
		return self._status

	@property
	def content_type(self) -> str:
		return self._type

	@property
	def headers(self) -> list:
		return self._headers

	@property
	def raw_response(self) -> callable:
		return self._raw
